/// <reference types="node" />
import * as events from 'events';
import ValueCategory from './value-category';
import { PlatformData, Message, RDL } from './protocol';
import { EntityMap } from '../utils/entity-utils';
import type Engine from '../engine';
import { DialogueTurn } from "../dataset-tools/parsers";
import AppExecutor from '../engine/apps/app_executor';
export interface ConversationOptions {
    nluServerUrl?: string;
    nlgServerUrl?: string;
    anonymous?: boolean;
    debug?: boolean;
    rng?: () => number;
    inactivityTimeout?: number;
    contextResetTimeout?: number;
    showWelcome?: boolean;
    deleteWhenInactive?: boolean;
    log?: boolean;
    dialogueFlags?: Record<string, boolean>;
}
interface Statistics {
    hit(key: string): void;
}
interface Context {
    code: string[];
    entities: EntityMap;
}
export interface ConversationDelegate {
    setHypothesis(hyp: string): void;
    setExpected(expect: string | null, ctx: Context): void;
    addMessage(msg: Message): Promise<void>;
}
export declare class DialogueTurnLog {
    private readonly _turn;
    private _done;
    constructor();
    get turn(): DialogueTurn;
    get done(): boolean;
    finish(): void;
    update(field: Exclude<keyof DialogueTurn, 'agent_timestamp' | 'user_timestamp'>, value: string): void;
}
export interface ConversationState {
    history: Message[];
    dialogueState: string;
}
/**
 * A single session of conversation in Genie.
 *
 * This object is responsible for maintaining the history of the conversation
 * to support clients reconnecting to the same conversation later, as well
 * as tracking connected clients and inactivity timeouts.
 *
 * The actual conversation logic is in {@link DialogueLoop}.
 */
export default class Conversation extends events.EventEmitter {
    private _engine;
    private _conversationId;
    private _locale;
    _: (x: string) => string;
    private _stats;
    private _options;
    private _debug;
    private _dialogueFlags;
    rng: () => number;
    private _loop;
    private _expecting;
    private _context;
    private _delegates;
    private _history;
    private _nextMsgId;
    private _inactivityTimeout;
    private _inactivityTimeoutSec;
    private _contextResetTimeout;
    private _contextResetTimeoutSec;
    private _log;
    constructor(engine: Engine, conversationId: string, options?: ConversationOptions);
    get isAnonymous(): boolean;
    get id(): string;
    get engine(): Engine;
    get stats(): Statistics;
    get history(): Message[];
    get inRecordingMode(): boolean;
    get dialogueFlags(): Record<string, boolean>;
    startRecording(): void;
    endRecording(): void;
    notify(app: AppExecutor, outputType: string, outputValue: Record<string, unknown>): void;
    notifyError(app: AppExecutor, error: Error): void;
    setExpected(expecting: ValueCategory | null, context: Context): void;
    start(state?: ConversationState): Promise<void>;
    stop(): Promise<void>;
    private _resetInactivityTimeout;
    addOutput(out: ConversationDelegate, replayHistory?: boolean): Promise<void>;
    removeOutput(out: ConversationDelegate): Promise<void>;
    private _callDelegates;
    setHypothesis(hypothesis: string): Promise<void>;
    sendAskSpecial(): Promise<void>;
    /**
     * Add a message to the conversation history.
     *
     * This method is exported to inject conversation history from outside.
     */
    addMessage(msg: Message): Promise<void>;
    /**
     * Extract the state from the conversation.
     *
     * This method is provided to save and restore the conversation state,
     * and transfer the conversation state between engines.
     */
    getState(): ConversationState;
    handleCommand(command: string, platformData?: PlatformData): Promise<void>;
    handleParsedCommand(root: any, title?: string, platformData?: PlatformData): Promise<void>;
    handleThingTalk(program: string, platformData?: PlatformData): Promise<void>;
    sendReply(message: string, icon: string | null): Promise<void>;
    sendMedia(mediaType: 'picture' | 'audio' | 'video', url: string, alt: string | undefined, icon: string | null): Promise<void>;
    sendRDL(rdl: RDL, icon: string | null): Promise<void>;
    sendSoundEffect(name: string, exclusive: boolean, icon: string | null): Promise<void>;
    sendChoice(idx: number, title: string): Promise<void>;
    sendButton(title: string, json: string): Promise<void>;
    sendLink(title: string, url: string, state: ConversationState): Promise<void>;
    sendNewProgram(program: {
        uniqueId: string;
        name: string;
        code: string;
        results: Array<Record<string, unknown>>;
        errors: string[];
        icon: string | null;
    }): Promise<void>;
    private get _lastDialogue();
    private get _lastTurn();
    appendNewDialogue(): void;
    dialogueFinished(): void;
    turnFinished(): void;
    appendNewTurn(turn: DialogueTurnLog): void;
    voteLast(vote: 'up' | 'down'): void;
    commentLast(comment: string): void;
    updateLog(field: Exclude<keyof DialogueTurn, 'agent_timestamp' | 'user_timestamp'>, value: string): void;
    saveLog(): Promise<void>;
    get log(): string | null;
}
export {};
